package org.foo.bytebuddyexample;

import static net.bytebuddy.matcher.ElementMatchers.named;

import java.io.IOException;
import java.lang.reflect.Constructor;
import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.util.Collections;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import net.bytebuddy.ByteBuddy;
import net.bytebuddy.NamingStrategy;
import net.bytebuddy.description.modifier.FieldManifestation;
import net.bytebuddy.description.modifier.SynchronizationState;
import net.bytebuddy.description.modifier.Visibility;
import net.bytebuddy.description.type.TypeDescription;
import net.bytebuddy.dynamic.DynamicType;
import net.bytebuddy.dynamic.Transformer;
import net.bytebuddy.dynamic.loading.ClassLoadingStrategy;
import net.bytebuddy.dynamic.scaffold.subclass.ConstructorStrategy;
import net.bytebuddy.implementation.MethodCall;
import net.bytebuddy.implementation.MethodDelegation;
import net.bytebuddy.implementation.SuperMethodCall;
import net.bytebuddy.matcher.ElementMatchers;

public class EntityModificationUtils {

  // The weird name is just to avoid collisions with other classes.
  private static final String MODIFIED_FIELDS_TRACKER_SUBCLASS_SUFFIX =
      "$Setter$Tracking$ByteBuddy$Subclass";

  // Define a constant suffix so we can just look at class name endings to determine if an entity
  // tracks modified fields.
  private static final NamingStrategy MODIFIED_FIELDS_TRACKER_SUBCLASS_NAMING_STRATEGY =
      new NamingStrategy() {
        @Override
        public String subclass(TypeDescription.Generic superClass) {
          return superClass.getTypeName() + MODIFIED_FIELDS_TRACKER_SUBCLASS_SUFFIX;
        }

        @Override
        public String redefine(TypeDescription typeDescription) {
          return "";
        }

        @Override
        public String rebase(TypeDescription typeDescription) {
          return "";
        }
      };

  // The weird name is just to avoid collisions with other fields.
  private static final String MODIFIED_FIELDS_TRACKER_FIELD_NAME = "_modified$Fields$Tracker";

  // The weird name is just to avoid collisions with other methods.
  private static final String INSTANCE_SAVE_METHOD_NAME = "_instance$Save$Method";

  // Given an Entity class, returns the corresponding ByteBuddy-generated subclass.
  static final Map<Class<?>, Class<?>> entityClassesToSetterTrackingEntityClasses =
      new ConcurrentHashMap<>();

  private static String fieldName(Method setter) {
    String fieldName = setter.getName().replaceFirst("set", "");
    return fieldName.substring(0, 1).toLowerCase() + fieldName.substring(1);
  }

  public static boolean tracksModifiedFields(Object entity) {
    return entity.getClass().getCanonicalName().endsWith(MODIFIED_FIELDS_TRACKER_SUBCLASS_SUFFIX);
  }

  // Gets the modified-fields tracker from a ByteBuddy-generated subclass. Because the subclass is
  // dynamically generated, so we have to use reflection to get it (there's no getter method
  // available to call directly).
  static Set<String> getModifiedFieldsTracker(Object entity) {
    try {
      Field modifiedFieldsTrackerField =
          entity.getClass().getDeclaredField(MODIFIED_FIELDS_TRACKER_FIELD_NAME);
      modifiedFieldsTrackerField.setAccessible(true);
      return (Set<String>) modifiedFieldsTrackerField.get(entity);
    } catch (Exception e) {
      throw new RuntimeException(e);
    }
  }

  public static void registerCalledSetter(Object entity, Method setter) {
    getModifiedFieldsTracker(entity).add(fieldName(setter));
  }

  public static Set<String> getModifiedFields(Object entity) {
    return Collections.unmodifiableSet(getModifiedFieldsTracker(entity));
  }

  static Constructor<?> getConstructor(Class<?> classToConstruct) {
    try {
      return classToConstruct.getConstructor();
    } catch (Exception e) {
      throw new IllegalStateException(
          "Unable to get public default constructor for " + classToConstruct);
    }
  }

  public static DynamicType.Unloaded<?> setterTrackingEntityType(Class<?> entityClass) {
    return new ByteBuddy()
        .with(MODIFIED_FIELDS_TRACKER_SUBCLASS_NAMING_STRATEGY)
        .subclass(entityClass, ConstructorStrategy.Default.NO_CONSTRUCTORS)
        .defineField(
            MODIFIED_FIELDS_TRACKER_FIELD_NAME,
            TypeDescription.Generic.Builder.parameterizedType(Set.class, String.class).build(),
            Visibility.PRIVATE,
            FieldManifestation.FINAL)
        .defineConstructor(Visibility.PUBLIC)
        .intercept(
            MethodCall.invoke(getConstructor(entityClass))
                .onSuper()
                .andThen(
                    MethodCall.construct(getConstructor(HashSet.class))
                        .setsField(named(MODIFIED_FIELDS_TRACKER_FIELD_NAME))))
        .method(ElementMatchers.isSetter())
        .intercept(MethodDelegation.to(SetterInterceptor.class).andThen(SuperMethodCall.INSTANCE))
        .transform(Transformer.ForMethod.withModifiers(SynchronizationState.SYNCHRONIZED))
        .defineMethod(INSTANCE_SAVE_METHOD_NAME, entityClass, Modifier.PUBLIC)
        .withParameters(Saver.class, Flags.class)
        .intercept(MethodDelegation.to(SaveInterceptor.class))
        .transform(Transformer.ForMethod.withModifiers(SynchronizationState.SYNCHRONIZED))
        .make();
  }

  public static Class<?> setterTrackingEntityClassImpl(Class<?> entityClass) {
    // Write it so we can inspect it. In production code this would be removed and this method would
    // just return setterTrackingEntityClassImpl(entityClass).
    // .load(entityClass.getClassLoader(), ClassLoadingStrategy.Default.INJECTION)
    // .getLoaded();
    DynamicType.Unloaded<?> setterTrackingEntityType = setterTrackingEntityType(entityClass);
    try {
      Files.write(
          Paths.get(System.getProperty("user.dir"), "build", "saved-subclass.class"),
          setterTrackingEntityType.getBytes());
    } catch (IOException e) {
      System.out.println("Unable to save ByteBuddy-generated subclass for inspection");
    }
    return setterTrackingEntityType
        .load(entityClass.getClassLoader(), ClassLoadingStrategy.Default.INJECTION)
        .getLoaded();
  }

  // Given an Entity class, returns the corresponding setter-tracking Entity class generated by
  // ByteBuddy.
  public static Class<?> setterTrackingEntityClass(Class<?> entityClass) {
    return entityClassesToSetterTrackingEntityClasses.computeIfAbsent(
        entityClass, v -> setterTrackingEntityClassImpl(entityClass));
  }

  // The Saver.saveImpl method is private, so we have to use reflection to get it and call it.
  private static Method getSaverSaveImplMethod() {
    try {
      Method saveImplMethod = Saver.class.getDeclaredMethod("saveImpl", Object.class, Flags.class);
      saveImplMethod.setAccessible(true);
      return saveImplMethod;
    } catch (Exception e) {
      throw new RuntimeException(e);
    }
  }

  private static final Method saverSaveImplMethod = getSaverSaveImplMethod();

  static <T> T saveImpl(T entity, Saver saver, Flags flags) {
    System.out.println("EntityModificationUtils.saveImpl called for ByteBuddy-generated subclass");
    Set<String> modifiedFieldsTracker = getModifiedFieldsTracker(entity);
    if (modifiedFieldsTracker.isEmpty()) {
      System.out.println("No fields are modified, returning");
      // Nothing to do, so just return the entity.
      return entity;
    }
    try {
      System.out.println("Has modified field(s), calling Saver.saveImpl method");
      T savedEntity = (T) saverSaveImplMethod.invoke(saver, entity, flags);
      // Clear the modified-fields tracker so that if another thread shares the same entity object,
      // only the first call to save does anything.
      System.out.println("Saved entity, clearing modified-fields tracker");
      getModifiedFieldsTracker(savedEntity).clear();
      System.out.println("Returning saved entity (which is the same object), and releasing lock");
      return savedEntity;
    } catch (InvocationTargetException | IllegalAccessException e) {
      throw new RuntimeException(e);
    }
  }

  // The ByteBuddy-generated instance-level synchronized save method is dynamically generated, so we
  // have to use reflection to get it and call it.
  private static <T> Method getInstanceLevelSynchronizedSaveMethod(T entity)
      throws NoSuchMethodException {
    Method saveMethod =
        entity.getClass().getDeclaredMethod(INSTANCE_SAVE_METHOD_NAME, Saver.class, Flags.class);
    saveMethod.setAccessible(true);
    return saveMethod;
  }

  public static <T> T save(T entity, Saver saver, Flags flags) {
    System.out.println(
        "Save called for ByteBuddy-generated subclass, delegating to ByteBuddy-generated instance-level synchronized save method");
    try {
      return (T) getInstanceLevelSynchronizedSaveMethod(entity).invoke(entity, saver, flags);
    } catch (InvocationTargetException | IllegalAccessException | NoSuchMethodException e) {
      throw new RuntimeException(e);
    }
  }
}
